import numpy as np
import pandas as pd
import copy
from sksurv.functions import StepFunction
from typing import Callable, Optional
from survlimepy.utils.step_function import transform_step_function


def interpolate_values(
    matrix: np.ndarray,
    unique_times_to_event: np.ndarray,
    model_output_times: np.ndarray,
) -> np.ndarray:
    """Transform an array of step functions to a matrix.

    Args:
        matrix (np.ndarray): matrix with values.
        unique_times_to_event (np.ndarray): unique times to event.
        model_output_times (np.ndarray): times considered by teh model.

    Returns:
        interpolated_matrix (np.ndarray): array containing the values.
    """
    interpolated_matrix = np.array(
        [
            np.interp(unique_times_to_event, model_output_times, matrix[i])
            for i in range(matrix.shape[0])
        ]
    )
    return interpolated_matrix


def validate_predicted_matrix(
    matrix: np.ndarray,
    individuals: Optional[np.ndarray] = None,
    expected_num_rows: Optional[int] = None,
    expected_num_cols: Optional[int] = None,
) -> None:
    """Validate the output.

    Args:
        matrix (np.ndarray): matrix to validate.
        individuals (Optional[np.ndarray]): individuals to be predicted.
        expected_num_rows (Optional[int]): expected number of rows.
        expected_num_cols (Optional[int]): expected number of columns.

    Returns:
        None.
    """
    total_rows = matrix.shape[0]
    total_cols = matrix.shape[1]
    are_nan_values = np.isnan(matrix).any()
    if are_nan_values:
        msg = "There are nan values produced by predict_fn function."
        if individuals is not None:
            idx = np.argwhere(np.isnan(matrix))
            example = str(individuals[idx[0][0], :])
            msg = f"{msg} Try to predict {example} array."
        raise ValueError(msg)
    if expected_num_rows:
        if total_rows != expected_num_rows:
            raise ValueError(
                f"The predicted function returns {total_rows} rows while expecting {expected_num_rows} rows."
            )
    if expected_num_cols:
        if total_cols != expected_num_cols:
            raise ValueError(
                f"The predicted function returns {total_cols} columns while expecting {expected_num_cols} columns."
            )

    return None


def predict_wrapper(
    predict_fn: Callable,
    data: np.ndarray,
    unique_times_to_event: np.ndarray,
    model_output_times: np.ndarray,
) -> np.ndarray:
    """Return the matrix with the values predicted.

    Args:
        predict_fn (Callable): function that computes cumulative hazard.
        data (np.ndarray): data to predict over.
        unique_times_to_event (np.ndarray): unique times to event.
        model_output_times (np.ndarray): output times of the bb model.

    Returns:
        predicted_values(np.ndarray): predicted values.
    """
    num_individuals = data.shape[0]
    number_unique_times = unique_times_to_event.shape[0]
    # Predict
    values_raw = predict_fn(data)
    # In case of a pd.DataFrame, force numpy
    if isinstance(values_raw, pd.DataFrame):
        values = values_raw.to_numpy()
    else:
        values = copy.deepcopy(values_raw)
    # If it is a numpy array
    if isinstance(values, np.ndarray):
        total_dim = values.ndim
        # If it is a matrix
        if total_dim == 2:
            total_cols = values.shape[1]
            # The produced number of columns matches with the number of unique times to event
            if total_cols == number_unique_times:
                predicted_values = values
            # The number of columns does not match with the number of unique times to event
            else:
                validate_predicted_matrix(
                    matrix=values,
                    individuals=data,
                    expected_num_rows=num_individuals,
                )
                predicted_values = interpolate_values(
                    matrix=values,
                    unique_times_to_event=unique_times_to_event,
                    model_output_times=model_output_times,
                )
        # It is an 1D array
        elif total_dim == 1:
            # It is a StepFunction
            if isinstance(values[0], StepFunction):
                predicted_values = transform_step_function(array_step_functions=values)
                if predicted_values.shape[1] != number_unique_times:
                    validate_predicted_matrix(
                        matrix=predicted_values,
                        individuals=data,
                        expected_num_rows=num_individuals,
                    )
                    predicted_values = interpolate_values(
                        matrix=predicted_values,
                        unique_times_to_event=unique_times_to_event,
                        model_output_times=model_output_times,
                    )
            else:
                raise NotImplemented("Unknown type of object.")
    else:
        raise TypeError("Unknown type of object.")

    validate_predicted_matrix(
        matrix=predicted_values,
        individuals=data,
        expected_num_rows=num_individuals,
        expected_num_cols=number_unique_times,
    )
    return predicted_values



import numpy as np


def transform_step_function(
    array_step_functions: np.array, return_column_vector: bool = False
) -> np.array:
    """Transform an array of step functions to a matrix.

    Args:
        array_step_functions (np.array): array of step functions.
        return_column_vector (bool): boolean indicating whether the value returned must be a column vector or not.

    Returns:
        matrix_values (np.array): array containing the values.
    """
    sample_value = array_step_functions[0]
    total_rows = len(array_step_functions)
    total_columns = len(sample_value.y)
    matrix_values = np.empty(shape=(total_rows, total_columns))
    for i, step_fn in enumerate(array_step_functions):
        matrix_values[i] = step_fn(step_fn.x)
    if return_column_vector:
        n_rows, n_cols = matrix_values.shape
        if n_rows == 1 and n_cols >= 1:
            matrix_values = matrix_values.T
    return matrix_values



import numpy as np
from functools import partial
from sksurv.linear_model import CoxPHSurvivalAnalysis
from sklearn.utils import check_random_state
from survlimepy import SurvLimeExplainer
from survlimepy.load_datasets import RandomSurvivalData
from survlimepy.utils.neighbours_generator import NeighboursGenerator
from survlimepy.load_datasets import Loader
from typing import List, Dict
import pandas as pd
import pytest


def test_shape_veterans_preprocessed() -> None:
    loader = Loader(dataset_name="veterans")
    x, _, _ = loader.load_data()
    assert x.shape == (137, 6)


def test_shape_udca_preprocessed() -> None:
    loader = Loader(dataset_name="udca")
    x, _, _ = loader.load_data()
    assert x.shape == (170, 4)


def test_shape_lung_preprocessed() -> None:
    loader = Loader(dataset_name="lung")
    x, _, _ = loader.load_data()
    assert x.shape == (226, 8)


def test_shape_pbc_preprocessed() -> None:
    loader = Loader(dataset_name="pbc")
    x, _, _ = loader.load_data()
    assert x.shape == (419, 17)


def test_shape_vetearns_computed_weights() -> None:
    loader = Loader(dataset_name="veterans")
    x, events, times = loader.load_data()
    train, test = loader.preprocess_datasets(x, events, times, random_seed=0)
    b = compute_weights(train, test)
    assert len(b) == 8


def test_shape_udca_computed_weights() -> None:
    loader = Loader(dataset_name="udca")
    x, events, times = loader.load_data()
    train, test = loader.preprocess_datasets(x, events, times, random_seed=0)
    b = compute_weights(train, test)
    assert len(b) == 4


def test_shape_lung_computed_weights() -> None:
    loader = Loader(dataset_name="lung")
    x, events, times = loader.load_data()
    train, test = loader.preprocess_datasets(x, events, times, random_seed=0)
    b = compute_weights(train, test)
    assert len(b) == 9


def test_norm_less_than_one() -> None:
    loader = Loader(dataset_name="veterans")
    x, events, times = loader.load_data()
    train, test = loader.preprocess_datasets(x, events, times, random_seed=0)
    with pytest.raises(ValueError):
        compute_weights(train, test, norm=0.5)


def test_categorical_features() -> None:
    n = 100
    random_state = check_random_state(2)
    data = {
        "col1": random_state.normal(size=n),
        "col2": random_state.normal(size=n),
        "col3": random_state.choice(a=["a", "b", "c"], size=n),
        "col4": random_state.choice(a=["d", "e", "f"], size=n),
    }

    df = pd.DataFrame(data)
    data_row = df.loc[0].to_numpy().reshape(1, -1)

    neighbours_generator = NeighboursGenerator(
        training_features=df,
        data_row=data_row,
        categorical_features=[2, 3],
        random_state=random_state,
    )

    neighbours = neighbours_generator.generate_neighbours(100)
    neighbours_first = neighbours[0, 2:4]
    expected_results = neighbours_first[0] == "a" and neighbours_first[1] == "f"
    assert expected_results == True


def test_num_rows() -> None:
    loader = Loader(dataset_name="lung")
    x, events, times = loader.load_data()
    train, test = loader.preprocess_datasets(x, events, times, random_seed=0)
    with pytest.raises(ValueError):
        compute_weights(train, test, mock_predict_fn=True)


def test_montecarlo_simulation() -> None:
    # Generate data
    n_points = 500
    true_coef = [1, 2]
    r = 1
    center = [0, 0]
    prob_event = 0.9
    lambda_weibull = 10 ** (-6)
    v_weibull = 2

    rsd = RandomSurvivalData(
        center=center,
        radius=r,
        coefficients=true_coef,
        prob_event=prob_event,
        lambda_weibull=lambda_weibull,
        v_weibull=v_weibull,
        time_cap=None,
        random_seed=90,
    )

    # Train
    X, time_to_event, delta = rsd.random_survival_data(num_points=n_points)
    z = [(d, t) for d, t in zip(delta, time_to_event)]
    y = np.array(z, dtype=[("delta", np.bool_), ("time_to_event", np.float32)])

    # Fit a Cox model
    cox = CoxPHSurvivalAnalysis()
    cox.fit(X, y)

    # User montecarlo
    data = np.array([[0, 0], [1, 1]])
    explainer = SurvLimeExplainer(
        training_features=X,
        training_events=[tp[0] for tp in y],
        training_times=[tp[1] for tp in y],
        model_output_times=cox.event_times_,
        random_state=10,
    )

    explanations = explainer.montecarlo_explanation(
        data=data,
        predict_fn=cox.predict_cumulative_hazard_function,
        num_samples=1,
        num_repetitions=1,
    )

    assert explanations.shape == data.shape


def test_value_error_center_coefficients(random_survidal_data) -> None:
    with pytest.raises(ValueError):
        RandomSurvivalData(
            center=[1, 2, 3],
            radius=random_survidal_data["radius"],
            coefficients=random_survidal_data["coefficients"],
            prob_event=random_survidal_data["prob_event"],
            lambda_weibull=random_survidal_data["lambda_weibull"],
            v_weibull=random_survidal_data["v_weibull"],
        )


def test_value_error_negative_lambda_weibull(random_survidal_data) -> None:
    with pytest.raises(ValueError):
        RandomSurvivalData(
            center=random_survidal_data["center"],
            radius=random_survidal_data["radius"],
            coefficients=random_survidal_data["coefficients"],
            prob_event=random_survidal_data["prob_event"],
            lambda_weibull=-1,
            v_weibull=random_survidal_data["v_weibull"],
        )


def test_value_error_negative_v_weibull(random_survidal_data) -> None:
    with pytest.raises(ValueError):
        RandomSurvivalData(
            center=random_survidal_data["center"],
            radius=random_survidal_data["radius"],
            coefficients=random_survidal_data["coefficients"],
            prob_event=random_survidal_data["prob_event"],
            lambda_weibull=random_survidal_data["lambda_weibull"],
            v_weibull=-1,
        )


def test_shape_spherical_data(random_survidal_data) -> None:
    num_points = 100
    center = random_survidal_data["center"]
    rsd = RandomSurvivalData(
        center=center,
        radius=random_survidal_data["radius"],
        coefficients=random_survidal_data["coefficients"],
        prob_event=random_survidal_data["prob_event"],
        lambda_weibull=random_survidal_data["lambda_weibull"],
        v_weibull=random_survidal_data["v_weibull"],
    )
    data = rsd.spherical_data(num_points=num_points)
    data_shape = data.shape
    assert data_shape == (num_points, len(center))


@pytest.fixture
def random_survidal_data() -> Dict:
    data = {
        "center": [0, 0],
        "radius": 1,
        "coefficients": [1, 2],
        "prob_event": 0.9,
        "lambda_weibull": 10 ** (-5),
        "v_weibull": 2,
    }
    return data


def compute_weights(
    train: np.array, test: np.array, norm: float = 2, mock_predict_fn: bool = False
) -> List[float]:
    model = CoxPHSurvivalAnalysis(alpha=0.0001)

    model.fit(train[0], train[1])
    events = [y[0] for y in train[1]]
    times = [y[1] for y in train[1]]

    times_to_fill = list(set([x[1] for x in train[1]]))
    times_to_fill.sort()

    explainer = SurvLimeExplainer(
        train[0],
        events,
        times,
        functional_norm=norm,
        model_output_times=model.event_times_,
    )

    num_pat = 1000
    test_point = test[0].iloc[0, :]

    predict_chf = partial(model.predict_cumulative_hazard_function, return_array=True)
    if mock_predict_fn:
        predict_chf = predict_chf_mocked(model.predict_cumulative_hazard_function)

    b = explainer.explain_instance(
        test_point,
        predict_fn=predict_chf,
        verbose=False,
        num_samples=num_pat,
    )
    return b


def predict_chf_mocked(predict_fn):
    def inner(X):
        total_rows = X.shape[1]
        pred_values = predict_fn(X=X, return_array=True)
        return pred_values[: (total_rows - 1), :]

    return inner



import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sksurv.util import Surv
from typing import List


def prepare_splits(
    x: pd.DataFrame,
    y: np.ndarray,
    split_sizes: List = [0.1, 0.5],
    random_state: int = 1,
) -> List[List, List]:
    """Creates a random split with given sizes and random seed.

    Args:
        x (pd.DataFrame): variable containing the dataset.
        y (np.ndarray): variable containing the target values.
        split_sizes (List): variable containing the percentage of the first and second split.
        random_state (int): int: seed of the random state to be used when splitting.

    Returns:
        X_x: pd.DataFrame: three different DataFrames with the splits' data.
        y_y: ----: three different ---- with the splits' targets.
    """
    X_train, X_test, y_train, y_test = train_test_split(
        x.copy(),
        y,
        test_size=split_sizes[0],
        random_state=random_state,
    )
    X_val, X_test, y_val, y_test = train_test_split(
        X_test.copy(), y_test, test_size=split_sizes[1], random_state=random_state
    )
    scaler = StandardScaler()

    X_train_processed = pd.DataFrame(
        data=scaler.fit_transform(X_train, y_train),
        columns=X_train.columns,
        index=X_train.index,
    )

    X_test_processed = pd.DataFrame(
        data=scaler.transform(X_test),
        columns=X_test.columns,
        index=X_test.index,
    )

    events_train = [x[0] for x in y_train]
    times_train = [x[1] for x in y_train]

    events_val = [x[0] for x in y_val]
    times_val = [x[1] for x in y_val]

    events_test = [x[0] for x in y_test]
    times_test = [x[1] for x in y_test]

    y_train = Surv.from_arrays(events_train, times_train)
    y_val = Surv.from_arrays(events_val, times_val)
    y_test = Surv.from_arrays(events_test, times_test)

    return X_train_processed, y_train, X_val, y_val, X_test_processed, y_test



from functools import partial
import logging
from copy import deepcopy
from typing import Callable, Tuple, Union, List, Literal, Optional
import numpy as np
import pandas as pd
import seaborn as sns
from tqdm import tqdm
from matplotlib import pyplot as plt
from sklearn.utils import check_random_state
from sksurv.functions import StepFunction
from survlimepy.utils.optimisation import OptFuncionMaker
from survlimepy.utils.neighbours_generator import NeighboursGenerator


class SurvLimeExplainer:
    """Look for the coefficient of a COX model."""

    def __init__(
        self,
        training_features: Union[np.ndarray, pd.DataFrame],
        training_events: Union[np.ndarray, pd.Series, List[Union[bool, float, int]]],
        training_times: Union[np.ndarray, pd.Series, List[Union[float, int]]],
        model_output_times: Optional[np.ndarray] = None,
        categorical_features: Optional[List[int]] = None,
        H0: Optional[Union[np.ndarray, List[float], StepFunction]] = None,
        kernel_width: Optional[float] = None,
        kernel_distance: str = "euclidean",
        kernel_fn: Optional[Callable] = None,
        functional_norm: Union[float, str] = 2,
        random_state: Optional[int] = None,
    ) -> None:
        """Init function.

        Args:
            training_features (Union[np.ndarray, pd.DataFrame]): data used to train the bb model.
            training_events (Union[np.ndarray, pd.Series, List[Union[bool, float, int]]]): training events indicator.
            training_times (Union[np.ndarray, pd.Series, List[Union[float, int]]]): training times to event.
            model_output_times (Optional[np.ndarray]): output times of the bb model.
            categorical_features (Optional[List[int]]): list of integers indicating the categorical features.
            H0 (Optional[Union[np.ndarray, List[float], StepFunction]]): baseline cumulative hazard.
            kernel_width (Optional[List[float]]): width of the kernel to be used for computing distances.
            kernel_distance (str): metric to be used for computing neighbours distance to the original point.
            kernel_fn (Optional[Callable]): kernel function to be used for computing distances.
            functional_norm (Optional[Union[float, str]]): functional norm to calculate the distance between the Cox model and the black box model.
            random_state (Optional[int]): number to be used for random seeds.

        Returns:
            None.
        """
        self.random_state = check_random_state(random_state)
        self.training_features = training_features
        self.training_events = training_events
        self.training_times = training_times
        self.categorical_features = categorical_features
        self.model_output_times = model_output_times
        self.computed_weights = None
        self.montecarlo_weights = None
        self.is_data_frame = isinstance(self.training_features, pd.DataFrame)
        if self.is_data_frame:
            self.feature_names = self.training_features.columns
        else:
            self.feature_names = [
                f"feature_{i}" for i in range(self.training_features.shape[1])
            ]
        self.H0 = H0

        if kernel_width is None:
            kernel_width = np.sqrt(self.training_features.shape[1]) * 0.75
        kernel_width = float(kernel_width)

        self.kernel_distance = kernel_distance

        if kernel_fn is None:

            def kernel_fn(d: np.ndarray, kernel_width: float) -> np.ndarray:
                return np.sqrt(np.exp(-(d**2) / kernel_width**2))

        self.kernel_fn = partial(kernel_fn, kernel_width=kernel_width)
        self.functional_norm = functional_norm

    def transform_data(
        self, data: Union[np.ndarray, pd.DataFrame]
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Transforms data according with what the model needs.

        Args:
            data (Union[np.ndarray, pd.DataFrame]): data to be transformed.

        Returns:
            data_transformed (Union[np.ndarray, pd.DataFrame]): transformed data.
        """
        if self.is_data_frame:
            data_transformed = pd.DataFrame(data, columns=self.feature_names)
        else:
            training_dtype = self.training_features.dtype
            data_transformed = np.copy(data)
            data_transformed = data_transformed.astype(training_dtype)
        return data_transformed

    def explain_instance(
        self,
        data_row: Union[List[float], np.ndarray, pd.Series],
        predict_fn: Callable,
        type_fn: Literal["survival", "cumulative"] = "cumulative",
        num_samples: int = 1000,
        max_difference_time_allowed: Optional[float] = None,
        max_hazard_value_allowed: Optional[float] = None,
        verbose: bool = False,
    ) -> np.ndarray:
        """Generates explanations for a prediction.

        Args:
            data_row (Union[List[float], np.ndarray, pd.Series]): data point to be explained.
            predict_fn (Callable): function that computes cumulative hazard.
            type_fn (Literal["survival", "cumulative"]): whether predict_fn is the cumulative hazard funtion or survival function.
            num_samples (int): number of neighbours to use.
            max_difference_time_allowed (Optional[float]): maximum difference between times allowed. If a difference exceeds this value, then max_difference_time_allowed will be used.
            max_hazard_value_allowed (Optional[float]): maximum hazard value allowed. If a prediction exceeds this value, then max_hazard_value_allows will be used.
            verbose (bool): whether or not to show cvxpy messages.

        Returns:
            cox_values (np.ndarray): obtained weights from the convex problem.
        """
        # To be used while plotting
        if isinstance(data_row, list):
            self.data_point = np.array(data_row).reshape(1, -1)
        elif isinstance(data_row, np.ndarray):
            total_dimensions_data_row = len(data_row.shape)
            total_rows = data_row.shape[0]
            if total_dimensions_data_row == 1:
                self.data_point = np.reshape(data_row, newshape=(1, -1))
            elif total_dimensions_data_row == 2:
                if total_rows > 1:
                    raise ValueError("data_point must contain a single row.")
                self.data_point = data_row
            else:
                raise ValueError("data_point must not have more than 2 dimensions.")
        elif isinstance(data_row, pd.Series):
            self.data_point = data_row.to_numpy().reshape(1, -1)
        else:
            raise ValueError(
                "data_point must be a list, a numpy array or a pandas Series."
            )

        # Generate neighbours
        neighbours_generator = NeighboursGenerator(
            training_features=self.training_features,
            data_row=self.data_point,
            categorical_features=self.categorical_features,
            random_state=self.random_state,
        )

        scaled_data = neighbours_generator.generate_neighbours(num_samples=num_samples)
        scaled_data_transformed = self.transform_data(data=scaled_data)

        # Solve optimisation problem
        opt_funcion_maker = OptFuncionMaker(
            training_events=self.training_events,
            training_times=self.training_times,
            neighbours=scaled_data,
            neighbours_transformed=scaled_data_transformed,
            num_samples=num_samples,
            data_point=self.data_point,
            kernel_distance=self.kernel_distance,
            kernel_fn=self.kernel_fn,
            predict_fn=predict_fn,
            type_fn=type_fn,
            functional_norm=self.functional_norm,
            model_output_times=self.model_output_times,
            H0=self.H0,
            max_difference_time_allowed=max_difference_time_allowed,
            max_hazard_value_allowed=max_hazard_value_allowed,
            verbose=verbose,
        )
        b = opt_funcion_maker.solve_problem()
        self.computed_weights = np.copy(b)
        return b

    def plot_weights(
        self,
        figsize: Tuple[int, int] = (10, 10),
        feature_names: Optional[List[str]] = None,
        scale_with_data_point: bool = False,
        figure_path: Optional[str] = None,
    ) -> None:
        # Create docstring of the function
        """Plots the weights of the computed COX model.

        Args:
            figsize (Tuple[int, int]): size of the figure.
            feature_names (Optional[List[str]]): names of the features.
            scale_with_data_point (bool): whether to perform the elementwise multiplication between the point to be explained and the coefficients.
            figure_path (Optional[str]): path to save the figure.

        Returns:
            None.
        """
        if self.computed_weights is None:
            raise ValueError(
                "SurvLIME weights not computed yet. Call explain_instance first before using this function."
            )
        else:
            are_there_any_nan = np.isnan(self.computed_weights).any()
            if are_there_any_nan:
                raise ValueError("Some of the coefficients contain nan values.")

        if feature_names is not None:
            if len(feature_names) != self.computed_weights.shape[0]:
                raise TypeError(
                    f"feature_names must have {self.computed_weights[0]} elements."
                )
        else:
            feature_names = self.feature_names

        if scale_with_data_point:
            weights = self.computed_weights * self.data_point
        else:
            weights = self.computed_weights

        _, ax = plt.subplots(figsize=figsize)

        # sort weights in descending order
        idx_sort = np.argsort(weights)[::-1]
        sorted_weights = weights[idx_sort]
        # sort feature names so that they match the sorted weights
        sorted_feature_names = [feature_names[i] for i in idx_sort]

        # divide the sorted weights and sorted feature names into positive and negative
        pos_weights = [w for w in sorted_weights if w > 0]
        pos_feature_names = [
            f for f, w in zip(sorted_feature_names, sorted_weights) if w > 0
        ]
        neg_weights = [w for w in sorted_weights if w < 0]
        neg_feature_names = [
            f for f, w in zip(sorted_feature_names, sorted_weights) if w < 0
        ]

        for label, weights_separated, palette in zip(
            [pos_feature_names, neg_feature_names],
            [pos_weights, neg_weights],
            ["Reds_r", "Blues"],
        ):
            data = pd.DataFrame({"features": label, "weights": weights_separated})
            ax.bar(
                "features",
                "weights",
                data=data,
                color=sns.color_palette(palette, n_colors=len(label)),
                label=label,
            )

        ax.set_xlabel("Features", fontsize=14)
        ax.set_ylabel("SurvLIME value", fontsize=14)
        ax.set_title("Feature importance", fontsize=16, fontweight="bold")

        ax.tick_params(axis="x", labelsize=14, rotation=90)
        ax.tick_params(axis="y", labelsize=14)

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

        if figure_path is not None:
            plt.savefig(figure_path, dpi=200, bbox_inches="tight")

        plt.show()

    def montecarlo_explanation(
        self,
        data: Union[pd.DataFrame, pd.Series, np.ndarray, List],
        predict_fn: Callable,
        type_fn: Literal["survival", "cumulative"] = "cumulative",
        num_samples: int = 1000,
        num_repetitions: int = 10,
        max_difference_time_allowed: Optional[float] = None,
        max_hazard_value_allowed: Optional[float] = None,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """Generates explanations for a prediction.

        Args:
            data (Union[pd.DataFrame, pd.Series, np.ndarray, List]): data points to be explained.
            predict_fn (Callable): function that computes cumulative hazard.
            type_fn (Literal["survival", "cumulative"]): whether predict_fn is the cumulative hazard funtion or survival function.
            num_samples (int): number of neighbours to use.
            num_repetitions (int): number of times to repeat the explanation.
            max_difference_time_allowed (Optional[float]): maximum difference between times allowed. If a difference exceeds this value, then max_difference_time_allowed will be used.
            max_hazard_value_allowed (Optional[float]): maximum hazard value allowed. If a prediction exceeds this value, then max_hazard_value_allows will be used.
            verbose (bool): whether or not to show cvxpy messages.

        Returns:
            montecarlo_explanation (pd.DataFrame): dataframe with the montecarlo explanation.
        """
        if isinstance(data, pd.DataFrame):
            data_transformed = data.to_numpy()
        elif isinstance(data, pd.Series):
            data_transformed = data.to_numpy().reshape(1, -1)
        elif isinstance(data, np.ndarray):
            n_dim = len(data.shape)
            if n_dim == 1:
                n_rows = 1
            else:
                n_rows = data.shape[0]
            data_transformed = np.reshape(data, newshape=(n_rows, -1))
        elif isinstance(data, list):
            data_transformed = np.array(data).reshape(1, -1)
        else:
            raise ValueError(
                "data must be a pandas DataFrame, a pandas Series, a numpy array or a list."
            )
        self.matrix_to_explain = deepcopy(data_transformed)
        all_solved = True
        total_rows = data_transformed.shape[0]
        total_cols = data_transformed.shape[1]
        weights = np.empty(shape=(total_rows, total_cols))
        for i_row in tqdm(range(total_rows)):
            current_row = data_transformed[i_row]
            weights_current_row = []
            i_rep = 0
            while i_rep < num_repetitions:
                # sample data point from the dataset
                b = self.explain_instance(
                    data_row=current_row,
                    predict_fn=predict_fn,
                    type_fn=type_fn,
                    num_samples=num_samples,
                    max_difference_time_allowed=max_difference_time_allowed,
                    max_hazard_value_allowed=max_hazard_value_allowed,
                    verbose=verbose,
                )
                are_there_any_nan_value = np.isnan(b).any()

                if are_there_any_nan_value:
                    all_solved = False
                else:
                    weights_current_row.append(b)
                    i_rep += 1

            weights_current_row = np.array(weights_current_row)
            mean_weight_row = np.mean(weights_current_row, axis=0)
            weights[i_row] = mean_weight_row

        if not all_solved:
            logging.warning(
                "There were some simulations without a solution. Try to run it with verbose=True."
            )
        self.montecarlo_weights = weights
        return weights

    def plot_montecarlo_weights(
        self,
        figsize: Tuple[int, int] = (10, 10),
        feature_names: Optional[List[str]] = None,
        scale_with_data_point: bool = False,
        figure_path: Optional[str] = None,
    ) -> None:
        """Generates explanations for a prediction.

        Args:
            figsize (Tuple[int, int]): size of the figure.
            feature_names Optional[List[str]]): names of the features.
            scale_with_data_point (bool): whether to perform the elementwise multiplication between the point to be explained and the coefficients.
            figure_path (Optional[str]): path to save the figure.

        Returns:
            None.
        """
        if self.montecarlo_weights is None:
            raise ValueError(
                "Monte-Carlo weights not computed yet. Call montecarlo_explanation first before using this function."
            )
        else:
            are_there_any_nan = np.isnan(self.montecarlo_weights).any()
            if are_there_any_nan:
                raise ValueError("Some of the coefficients contain nan values.")

        total_cols = self.montecarlo_weights.shape[1]

        if feature_names is not None:
            if len(feature_names) != total_cols:
                raise TypeError(f"feature_names must have {total_cols} elements.")
            col_names = feature_names
        else:
            col_names = self.feature_names

        if scale_with_data_point:
            scaled_data = np.multiply(self.montecarlo_weights, self.matrix_to_explain)

        else:
            scaled_data = self.montecarlo_weights
        data = pd.DataFrame(data=scaled_data, columns=col_names)

        sns.set()
        median_up = {}
        median_down = {}
        threshold = 0
        for (columnName, columnData) in data.items():
            median_value = np.median(columnData)
            if median_value > threshold:
                median_up[columnName] = median_value
            else:
                median_down[columnName] = median_value

        median_up = dict(
            sorted(median_up.items(), key=lambda item: item[1], reverse=True)
        )
        median_down = dict(
            sorted(median_down.items(), key=lambda item: item[1], reverse=True)
        )
        pal_up = sns.color_palette("Reds_r", n_colors=len(median_up))
        pal_down = sns.color_palette("Blues", n_colors=len(median_down))
        colors_up = {key: val for key, val in zip(median_up.keys(), pal_up)}
        colors_down = {key: val for key, val in zip(median_down.keys(), pal_down)}
        custom_pal = {**colors_up, **colors_down}
        data_reindex = data.reindex(columns=custom_pal.keys())
        data_melt = pd.melt(data_reindex)

        _, ax = plt.subplots(figsize=figsize)
        ax.tick_params(labelrotation=90)
        p = sns.boxenplot(
            x="variable",
            y="value",
            data=data_melt,
            palette=custom_pal,
            ax=ax,
        )
        ax.tick_params(labelrotation=90)
        p.set_xlabel("Features", fontsize=14)
        p.set_ylabel("SurvLIME value", fontsize=14)
        p.yaxis.grid(True)
        p.xaxis.grid(True)

        p.set_title("Feature importance", fontsize=16, fontweight="bold")

        plt.xticks(fontsize=16, rotation=90)
        plt.yticks(fontsize=14, rotation=0)

        if figure_path is not None:
            plt.savefig(figure_path, dpi=200, bbox_inches="tight")

        plt.show()



from typing import Union, Callable, Optional, Literal, List
from sklearn.metrics import pairwise_distances
from survlimepy.utils.predict import predict_wrapper
from sksurv.nonparametric import nelson_aalen_estimator
from sksurv.functions import StepFunction
import numpy as np
import cvxpy as cp
import pandas as pd
import logging
from survlimepy.utils.step_function import transform_step_function


class OptFuncionMaker:
    """This class manages the optimisation process."""

    def __init__(
        self,
        training_events: Union[np.ndarray, pd.Series, List[Union[bool, float, int]]],
        training_times: Union[np.ndarray, pd.Series, List[Union[float, int]]],
        neighbours: np.ndarray,
        neighbours_transformed: Union[np.ndarray, pd.DataFrame],
        num_samples: int,
        data_point: np.ndarray,
        kernel_distance: str,
        kernel_fn: Callable,
        predict_fn: Callable,
        type_fn: Literal["survival", "cumulative"],
        functional_norm: Union[float, str],
        model_output_times: Optional[np.ndarray] = None,
        H0: Optional[Union[np.ndarray, List[float], StepFunction]] = None,
        max_difference_time_allowed: Optional[float] = None,
        max_hazard_value_allowed: Optional[float] = None,
        verbose: bool = True,
    ) -> None:
        """Init function.

        Args:
            training_events (Union[np.ndarray, pd.Series, List[Union[bool, float, int]]]): training events indicator.
            training_times (Union[np.ndarray, pd.Series, List[Union[float, int]]]): training times to event.
            neighbours (np.ndarray): neighbours (num_samples x features).
            neighbours_transformed (Union[np.ndarray, pd.DataFrame]): neighbours in the appropriate format to use the prediction function.
            num_samples (int): number of neighbours to use.
            data_point (np.ndarray): data point to be explained.
            kernel_distance (str): metric to be used for computing neighbours distance to the original point.
            kernel_fn (Callable): kernel function to be used for computing distances.
            predict_fn (Callable): function that computes cumulative hazard.
            type_fn (Literal["survival", "cumulative"]): whether predict_fn is the cumulative hazard funtion or survival function.
            functional_norm (Union[float, str]): functional norm to calculate the distance between the Cox model and the black box model.
            model_output_times (Optional[np.ndarray]): output times of the bb model.
            H0 (Optional[Union[np.ndarray, List[float], StepFunction]]): baseline cumulative hazard.
            max_difference_time_allowed (Optional[float]): maximum difference between times allowed. If a difference exceeds this value, then max_difference_time_allowed will be used.
            max_hazard_value_allowed (Optional[float]): maximum hazard value allowed. If a prediction exceeds this value, then max_hazard_value_allows will be used.
            verbose (bool): whether or not to show cvxpy messages.

        Returns:
            None.
        """
        self.validate_events_times(training_events)
        self.validate_events_times(training_times)
        if isinstance(training_events, pd.Series):
            self.training_events = training_events.to_numpy()
        else:
            self.training_events = training_events
        if isinstance(training_times, pd.Series):
            self.training_times = training_times.to_numpy()
        else:
            self.training_times = training_times
        self.neighbours = neighbours
        self.neighbours_transformed = neighbours_transformed
        self.num_samples = num_samples
        self.data_point = data_point
        self.kernel_distance = kernel_distance
        self.kernel_fn = kernel_fn
        self.predict_fn = predict_fn

        if type_fn not in ["survival", "cumulative"]:
            raise ValueError("type_fn must be either survival or cumulative string.")
        self.type_fn = type_fn

        self.unique_times_to_event = np.sort(np.unique(self.training_times))
        self.m = self.unique_times_to_event.shape[0]

        if model_output_times is None:
            self.model_output_times = np.copy(self.unique_times_to_event)
        else:
            self.model_output_times = model_output_times

        if (
            isinstance(functional_norm, float) or isinstance(functional_norm, int)
        ) and functional_norm < 1:
            raise ValueError(
                f"functional_norm should be greater than 1, given value {functional_norm}."
            )
        elif isinstance(functional_norm, str) and functional_norm != "inf":
            raise ValueError(f"Invalid string for functional_norm. It must be \"inf\".")
        if not (
            isinstance(functional_norm, float)
            or isinstance(functional_norm, int)
            or isinstance(functional_norm, str)
        ):
            raise TypeError(f"functional_norm must be int, float or str type.")
        self.functional_norm = functional_norm

        if H0 is None:
            self.H0 = self.compute_nelson_aalen_estimator(
                self.training_events, self.training_times
            )
        else:
            if isinstance(H0, list):
                self.H0 = np.array(H0).reshape(-1, 1)
            elif isinstance(H0, np.ndarray):
                total_dimensions_H0 = len(H0.shape)
                if total_dimensions_H0 == 1:
                    self.H0 = np.reshape(H0, newshape=(-1, 1))
                elif total_dimensions_H0 == 2:
                    if self.H0.shape[1] > 1:
                        raise ValueError("H0 must contain a single column.")
                    self.H0 = H0
                else:
                    raise ValueError("H0 must be an array of maximum 2 dimensions.")
            elif isinstance(H0, StepFunction):
                self.H0 = transform_step_function(
                    array_step_functions=np.array([H0]), return_column_vector=True
                )
            else:
                raise ValueError(
                    "H0 must be a list, a numpy array or a sksurv StepFunction."
                )

        if max_difference_time_allowed is not None:
            if max_difference_time_allowed < 0:
                raise ValueError("max_difference_time_allowed must be positive.")
        self.max_difference_time_allowed = max_difference_time_allowed

        if max_hazard_value_allowed is not None:
            if max_hazard_value_allowed < 0:
                raise ValueError("max_hazard_value_allowed must be positive.")
        self.max_hazard_value_allowed = max_hazard_value_allowed

        self.verbose = verbose

        if self.H0.shape[0] != self.m:
            raise ValueError(f"H0 must have {self.m} rows/elements.")
        if self.H0.shape[1] != 1:
            raise ValueError("H0 must have 1 column.")

        self.limit_H_warning = 500
        self.epsilon = 10 ** (-6)

    @staticmethod
    def validate_events_times(vector):
        if not (
            isinstance(vector, list)
            or isinstance(vector, np.ndarray)
            or isinstance(vector, pd.Series)
        ):
            raise TypeError(
                "Both training_events and training_times must be a list, a numpy arroy or a pandas Series."
            )
        if isinstance(vector, np.ndarray) and len(vector) > 1:
            raise TypeError(
                "Both training_events and training_times must be 1D numpy arrays"
            )
        return None

    @staticmethod
    def compute_nelson_aalen_estimator(
        event: np.ndarray, time: np.ndarray
    ) -> np.ndarray:
        """Compute Nelson-Aalen estimator.

        Args:
            event (np.ndarray): array of events.
            time (np.ndarray): array of times.

        Returns:
            H0 (np.ndarray): the Nelson-Aalen estimator
        """
        nelson_aalen = nelson_aalen_estimator(event, time)
        H0 = nelson_aalen[1]
        m = H0.shape[0]
        H0 = np.reshape(H0, newshape=(m, 1))
        return H0

    def get_weights(self):
        """Compute the weights of each individual."""
        # Compute weights.
        distances = pairwise_distances(
            self.neighbours, self.data_point, metric=self.kernel_distance
        ).ravel()
        weights = self.kernel_fn(distances)
        w = np.reshape(weights, newshape=(self.num_samples, 1))
        return w

    def get_predictions(self) -> np.ndarray:
        """Compute the prediction for each neighbour."""
        # Compute predictions
        FN_pred = predict_wrapper(
            predict_fn=self.predict_fn,
            data=self.neighbours_transformed,
            unique_times_to_event=self.unique_times_to_event,
            model_output_times=self.model_output_times,
        )
        if self.type_fn == "survival":
            H_score = -np.log(FN_pred + self.epsilon)
        else:
            H_score = np.copy(FN_pred)
        max_H_score = np.max(H_score)
        if self.max_hazard_value_allowed is None and max_H_score > self.limit_H_warning:
            logging.warning(
                f"The prediction function produces extreme values: {max_H_score}. In terms of survival, Pr(Survival) is almost 0. Try to set max_hazard_value_allowed parameter to clip this value."
            )
        if self.max_hazard_value_allowed is not None:
            H_score = np.clip(
                a=H_score, a_min=None, a_max=self.max_hazard_value_allowed
            )
        H = np.reshape(np.array(H_score), newshape=(self.num_samples, self.m))
        return H

    def get_delta_t(self) -> np.ndarray:
        """Compute the vector of delta times."""
        t = np.empty(shape=(self.m + 1, 1))
        t[: self.m, 0] = self.unique_times_to_event
        t[self.m, 0] = t[self.m - 1, 0] + self.epsilon
        delta_t = [t[i + 1, 0] - t[i, 0] for i in range(self.m)]
        delta_t = np.reshape(np.array(delta_t), newshape=(self.m, 1))
        if self.max_difference_time_allowed is not None:
            delta_t = np.clip(
                a=delta_t, a_min=None, a_max=self.max_difference_time_allowed
            )
        return delta_t

    def compute_norm(self, matrix: np.ndarray) -> np.ndarray:
        """Compute the norm of a given matrix.

        Args:
            matrix (np.ndarray): the matrix over which to compute the norm.

        Returns:
            matrix_norm (np.ndarray): The norm of the matrix
        """
        if isinstance(self.functional_norm, str):
            matrix_norm = cp.norm(matrix, "inf")
        elif self.functional_norm == 1:
            matrix_norm = cp.abs(matrix)
        elif self.functional_norm % 2 != 0:
            matrix_abs = cp.abs(matrix)
            matrix_norm = cp.power(matrix_abs, self.functional_norm)
        else:
            matrix_norm = cp.power(matrix, self.functional_norm)
        return matrix_norm

    def solve_problem(self) -> np.ndarray:
        """Solve the optimisation problem.

        Args:
            None.

        Returns:
            cox_coefficients (np.ndarray): The result of the optimisation problem.
        """
        # Varible to look for
        num_features = self.neighbours.shape[1]
        cox_coefficients = np.full(shape=num_features, fill_value=np.nan)
        b = cp.Variable((num_features, 1))

        # Get predictions
        H = self.get_predictions()
        LnH = np.log(H + self.epsilon)

        # Compute the log correction
        log_correction = np.divide(H, np.log(H + self.epsilon))

        # Log of baseline cumulative hazard
        LnH0 = np.log(self.H0 + self.epsilon)

        # Weights of individuals
        w = self.get_weights()

        # Time differences
        delta_t = self.get_delta_t()

        # Matrices to produce the proper sizes
        ones_N = np.ones(shape=(self.num_samples, 1))
        ones_m_1 = np.ones(shape=(self.m, 1))
        B = np.dot(ones_N, LnH0.T)
        C = LnH - B
        Z = self.neighbours @ b
        D = Z @ ones_m_1.T
        E = C - D
        E_norm = self.compute_norm(matrix=E)
        V_sq = cp.square(log_correction)
        F = cp.multiply(E_norm, V_sq)
        G = F @ delta_t
        funct = G.T @ w

        # Solving the problem
        try:
            objective = cp.Minimize(funct)
            prob = cp.Problem(objective)
            prob.solve(solver="OSQP", verbose=self.verbose, eps_abs=1e-3, eps_rel=1e-3)
            if prob.status not in ["infeasible", "unbounded"]:
                cox_coefficients = b.value[:, 0]
            else:
                logging.warning(f"The status of the problem is {prob.status}.")
        except (
            cp.error.DCPError,
            cp.error.DGPError,
            cp.error.DPPError,
            cp.error.SolverError,
        ) as e:
            msg = f"{str(e)}\n Returning an array full of nan values."
            logging.warning(msg)
        return cox_coefficients



from typing import List, Optional
import os
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sksurv.util import Surv


data_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "datasets")
veteran_path = os.path.join(data_path, "veteran.csv")
udca_path = os.path.join(data_path, "udca_dataset.csv")
pbc_path = os.path.join(data_path, "pbc_dataset.csv")
lung_path = os.path.join(data_path, "lung_dataset.csv")
heart_path = os.path.join(data_path, "heart_dataset.csv")


class Loader:
    def __init__(self, dataset_name: str = "veterans"):

        if dataset_name == "veterans":
            self.feature_columns = [
                "celltype",
                "trt",
                "karno",
                "diagtime",
                "age",
                "prior",
            ]
            self.categorical_columns = ["celltype"]
            self.df = pd.read_csv(veteran_path)
            # if value in prior is different from 0 change to 1
            self.df["prior"] = [1 if x != 0 else 0 for x in self.df["prior"]]
            # substract 1 from variable trt
            self.df["trt"] = [x - 1 for x in self.df["trt"]]
        elif dataset_name == "udca":
            self.feature_columns = ["bili", "stage", "riskscore", "trt"]
            self.categorical_columns = []
            self.df = pd.read_csv(udca_path)
        elif dataset_name == "pbc":
            self.feature_columns = [
                "age",
                "bili",
                "chol",
                "albumin",
                "ast",
                "ascites",
                "copper",
                "alk.phos",
                "trig",
                "platelet",
                "protime",
                "trt",
                "sex",
                "hepato",
                "spiders",
                "edema",
                "stage",
            ]
            self.categorical_columns = ["edema", "stage"]
            self.df = pd.read_csv(pbc_path)
            self.df["sex"] = [1 if x == "f" else 0 for x in self.df["sex"]]
        elif dataset_name == "lung":
            self.feature_columns = [
                "inst",
                "age",
                "sex",
                "ph.ecog",
                "ph.karno",
                "pat.karno",
                "meal.cal",
                "wt.loss",
            ]
            self.categorical_columns = ["ph.ecog"]
            self.df = pd.read_csv(lung_path)
            # Delete the row with ph.ecog = 3.0
            self.df = self.df[self.df["ph.ecog"] != 3.0]
            # delete the row with ph.ecog = nan
            self.df = self.df[self.df["ph.ecog"].notna()]
            # substract 1 to each value of the status column
            self.df["status"] = [x - 1 for x in self.df["status"]]

        elif dataset_name == "synthetic":
            ## TODO
            pass
        elif dataset_name == "heart":
            self.feature_columns = [
                "age",
                "anaemia",
                "creatinine_phosphokinase",
                "diabetes",
                "ejection_fraction",
                "high_blood_pressure",
                "platelets",
                "serum_creatinine",
                "serum_sodium",
                "sex",
                "smoking",
            ]
            self.categorical_columns = []
            self.df = pd.read_csv(heart_path)
        else:
            raise AssertionError(
                f"The give name {dataset_name} was not found in [veterans, udca, pbc, lung]."
            )

    def load_data(self) -> list([pd.DataFrame, np.ndarray]):
        """
        Loads a survival dataset.

        Returns:
        x (pd.DataFrame): unprocessed features.
        y (np.ndarray): tuples with (status, time).
        """
        self.df["status"] = [True if x == 1 else False for x in self.df["status"]]
        self.df["y"] = [
            (x, y) for x, y in zip(self.df["status"], self.df["time"])
        ]  # Needed for sksurv
        y = self.df.pop("y").to_numpy()
        events = [x[0] for x in y]
        times = [x[1] for x in y]
        x = self.df[self.feature_columns]

        x = x.fillna(value=x.median(numeric_only=True))

        return x, events, times

    def preprocess_datasets(
        self,
        x: pd.DataFrame,
        events: list,
        times: list,
        standarize: bool = True,
        random_seed: int = 0,
    ) -> list([pd.DataFrame, np.ndarray]):
        """
        Preprocesses the data to be used as model input.

        For now it only converts categorical features to OHE and
        standarizes the data.
        """
        # Deal with categorical features
        x_pre = x.copy()
        for cat_feat in self.categorical_columns:
            names = [cat_feat + "_" + str(value) for value in x_pre[cat_feat].unique()]
            df_dummy_i = pd.get_dummies(
                x_pre[cat_feat], drop_first=True, prefix=cat_feat
            )
            new_names = df_dummy_i.columns
            # x_pre[names[1:]] = pd.get_dummies(x_pre[cat_feat], drop_first=True)
            x_pre[new_names] = df_dummy_i
            x_pre.drop(cat_feat, inplace=True, axis=1)

        # Then convert the data and the features to three splits
        # and standarize them
        y = Surv.from_arrays(events, times)
        X_train, X_test, y_train, y_test = train_test_split(
            x_pre.copy(), y, test_size=0.10, random_state=random_seed
        )
        # X_val, X_test, y_val, y_test = train_test_split(
        #    X_test.copy(), y_test, test_size=0.5, random_state=random_seed
        # )

        if standarize:
            scaler = StandardScaler()
            X_train = pd.DataFrame(
                data=scaler.fit_transform(X_train, y_train),
                columns=X_train.columns,
                index=X_train.index,
            )

            #    X_val = pd.DataFrame(
            #        data=scaler.transform(X_val),
            #        columns=X_val.columns,
            #        index=X_val.index,
            #    )

            X_test = pd.DataFrame(
                data=scaler.transform(X_test),
                columns=X_test.columns,
                index=X_test.index,
            )

        return [X_train, y_train], [X_test, y_test]  # [X_val, y_val]


class RandomSurvivalData:
    """Generate spherical random survival data."""

    def __init__(
        self,
        center: List[float],
        radius: float,
        coefficients: List[float],
        prob_event: float,
        lambda_weibull: float,
        v_weibull: float,
        time_cap: float = 2000,
        random_seed: Optional[int] = None,
    ) -> None:
        """Init.
        Args:
            center (float): center of the sphere.
            radius (float): radius of the sphere.
            coefficients (list): value of the covariates.
            prob_event (float): probability of an event occurring.
            lambda_weibull (float): scale parameter of a Weibull distribution.
            v_weibull (float): shape parameter of a Weibull distribution.
            time_cap (float): if the time is greater than time_cap, then time_cap will be used
            random_seed (Optional[int]):  number to be used for random seeds.

        Returns:
            None.
        """
        if len(center) != len(coefficients):
            raise ValueError(
                "length of center and length of coefficients must be equal."
            )
        if prob_event <= 0:
            raise ValueError("prob_event must be greater than 0.")
        if prob_event >= 1:
            raise ValueError("prob_event must be less than 0.")
        if lambda_weibull <= 0:
            raise ValueError("lambda_weibull must be greater than 0.")
        if v_weibull <= 0:
            raise ValueError("v_weibull must be greater than 0.")
        self.center = center
        self.radius = radius
        self.coefficients = coefficients
        self.prob_event = prob_event
        self.lambda_weibull = lambda_weibull
        self.v_weibull = v_weibull
        self.random_state = np.random.default_rng(random_seed)
        self.time_cap = time_cap

    def spherical_data(self, num_points: int) -> np.ndarray:
        """Generates random data in the p-dimensional sphere (covariates).

        Args:
            num_points (int): number of individuals to generate.

        Returns:
            np.ndarray: matrix with num_points rows and p columns, where p is the dimension of the space.
        """
        center = self.center
        radius = self.radius

        n_dim = len(center)
        zero_list = [0 for _ in range(n_dim)]

        # Uniform data in range [0, 1]
        u = self.random_state.uniform(low=0.0, high=1.0, size=(num_points, 1))
        uniform_rad_root = u ** (1 / n_dim)
        uniform_rad = radius * uniform_rad_root

        # Multivariate normal distribution
        I = np.identity(n_dim)
        X = self.random_state.multivariate_normal(
            mean=zero_list, cov=I, size=num_points
        )

        # Standarisation
        X_sqr = np.sum(np.square(X), axis=1, keepdims=True)
        norm_X = np.sqrt(X_sqr)
        X_unit = X / norm_X
        sphere_data = uniform_rad * X_unit
        X_location = sphere_data + center

        return X_location

    def survival_times(self, num_points: int, X: np.ndarray) -> np.ndarray:
        """Generates survival times following a Weibull distribution.

        Args:
            num_points (int):  number of individuals to generate.
            X (np.ndarray): matrix with num_points rows and p columns, where p is the dimension of the space.

        Returns:
            np.ndarray: a column vector containing the survival times.
        """
        u = self.random_state.uniform(size=(num_points, 1))
        lamba_val = self.lambda_weibull
        v = self.v_weibull
        b = np.reshape(self.coefficients, newshape=(len(self.coefficients), 1))
        num = -np.log(u)
        den = lamba_val * np.exp(np.dot(X, b))
        # Use a Weibull distribution
        time_to_event = (num / den) ** (1 / v)
        if self.time_cap:
            time_to_event = np.where(
                time_to_event > self.time_cap, self.time_cap, time_to_event
            )
        if len(time_to_event.shape) == 2:
            time_to_event = time_to_event[:, 0]
        return time_to_event

    def random_event(self, num_points: int) -> np.ndarray:
        """Generates random events following a binomial distributiom with probabilty `prob_event`.

        Args:
            num_points (int):  number of individuals to generate.

        Returns:
            np.ndarray: a column vector containing the random events.
        """
        prob_event = self.prob_event
        return np.where(
            self.random_state.uniform(size=num_points) <= prob_event, True, False
        )

    def random_survival_data(self, num_points: int) -> tuple:
        """Generates random survival data.

        Args:
            num_points (int):  number of individuals to generate.

        Returns:
            tuple: (X, time_to_event, delta), where:
                - X: matrix with num_points rows and p columns, where p is the dimension of the space.
                - time_to_event: a column vector containing the survival times.
                - delta: a column vector containing the random events.
        """
        # Get spherical data
        X = self.spherical_data(num_points=num_points)

        # Get random survival time
        time_to_event = self.survival_times(num_points=num_points, X=X)

        # Get event variable
        delta = self.random_event(num_points=num_points)

        return (X, time_to_event, delta)



import numpy as np
import pandas as pd
from typing import List, Dict, Union, Optional


class NeighboursGenerator:
    """Manages the process of obtaining the Neighbours for a given row"""

    def __init__(
        self,
        training_features: Union[np.ndarray, pd.DataFrame],
        data_row: np.ndarray,
        categorical_features: Optional[List[int]] = None,
        random_state: Optional[int] = None,
    ) -> None:
        """Init function.

        Args:
            training_features (Union[np.ndarray, pd.DataFrame]): data used to train the bb model.
            data_row (np.ndarray): data point to be explained of shape (1 x features).
            categorical_features (Optional[List[int]]): list of integeter indicating the categorical features.
            random_state (Optional[int]): number to be used for random seeds.
        Returns:

            None.
        """
        if isinstance(training_features, pd.DataFrame):
            self.training_features = training_features.to_numpy()
        elif isinstance(training_features, list):
            self.training_features = np.array(training_features)
        else:
            self.training_features = training_features

        self.data_row = data_row
        self.total_features = self.training_features.shape[1]

        if categorical_features is None:
            self.cat_features = []
        else:
            self.cat_features = categorical_features

        self.cont_features = [
            i for i in range(self.total_features) if i not in self.cat_features
        ]
        self.total_cat_features = len(self.cat_features)
        self.total_cont_features = len(self.cont_features)

        if len(self.cat_features) == 0:
            self.type_features = "continuous"
        elif len(self.cont_features) == 0:
            self.type_features = "categorical"
        else:
            self.type_features = "mixed"

        self.random_state = random_state

    @staticmethod
    def to_dict(keys, values):
        return {key: value for key, value in zip(keys, values)}

    def estimate_distribution_categorical_features(self) -> Dict:
        """Estimates the distribution for each categorical variable.

        Args:
            None.

        Returns:
            distribution (dict): a dictionary containing the distribution for each categorical variable.

        """
        distribution = {}
        total = self.training_features.shape[0]
        if self.type_features in ["mixed", "categorical"]:
            for idx_feature in self.cat_features:
                feautre_values = self.training_features[:, idx_feature]
                unique, count = np.unique(feautre_values, return_counts=True)
                distribution[idx_feature] = self.to_dict(unique, count / total)
        return distribution

    def generate_cont_neighbours(self, num_samples: int) -> np.ndarray:
        """Generates a neighborhood around a prediction for continuous features.

        Args:
            num_samples (int): number of neighbours to generate.

        Returns:
            data (np.ndarray): original data point and neighbours with shape (num_samples x features).
        """
        # Get continuous features
        training_features_cont = self.training_features[:, self.cont_features]

        # Estimate the variance for continuous features
        sd_value = np.nanstd(training_features_cont, axis=0, dtype=np.float32)

        # Generate neighbours
        neighbours = self.random_state.normal(
            0, 1, size=(num_samples, self.total_cont_features)
        )
        data_row_cont = self.data_row[0][self.cont_features].astype(neighbours.dtype)
        neighbours *= sd_value
        neighbours += data_row_cont

        return neighbours

    def generate_cat_neighbours(self, num_samples: int) -> np.ndarray:
        """Generates a neighborhood around a prediction for continuous features.

        Args:
            num_samples (int): number of neighbours to generate.

        Returns:
            data (np.ndarray): original data point and neighbours with shape (num_samples x features).
        """
        # Generate distribution for categorical features
        cat_distribution = self.estimate_distribution_categorical_features()
        neighbours_list = []

        # Generate neighbours
        i = 0
        for _, feat_distribution in cat_distribution.items():
            feat_values = list(feat_distribution.keys())
            probability_values = list(feat_distribution.values())
            sample = self.random_state.choice(
                feat_values, size=num_samples, p=probability_values
            )
            neighbours_list.append(sample)
            i += 1

        neighbours = np.array(neighbours_list)
        neighbours = neighbours.T
        return neighbours

    def generate_neighbours(self, num_samples: int) -> np.ndarray:
        """Generates a neighborhood around a prediction.

        Args:
            num_samples (int): number of neighbours to generate.

        Returns:
            data (np.ndarray): original data point and neighbours with shape (num_samples x features).
        """
        # Generate neighbours for continuous features
        if self.type_features in ["continuous", "mixed"]:
            X_neigh_cont = self.generate_cont_neighbours(num_samples=num_samples)

        # Generate neighbours for categorical features
        if self.type_features in ["categorical", "mixed"]:
            X_neigh_cat = self.generate_cat_neighbours(num_samples=num_samples)

        # Create neighbours
        if self.type_features == "continuous":
            neighbours = np.copy(X_neigh_cont)
        elif self.type_features == "categorical":
            neighbours = np.copy(X_neigh_cat)
        else:
            neighbours = np.concatenate(
                (X_neigh_cont, X_neigh_cat), axis=1, dtype=object
            )

        return neighbours
